/*
 * Created on Apr 11, 2005
 *
 */
package com.sunlabs.archsim.fleet.HornAndFunnel;

import java.util.Map;
import java.util.HashMap;


import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 *
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 * 
 * @author ib27688
 *
 * Model of a Fleet destination Horn.
 * Interface: in, ack in on the input side,
 * out[i], ackOut[i] on the outputr side, where
 * i ranges from 0 to (2^depth)-1.
 */
public class DestinationHorn extends Component {


	
	/** the number of internal brancing levels */
	int depth;
	
	/** the number of out ports - on the output side */
	int numOutputs;
	
	final int branchFactor= 2;
	
	DestinationHornBranchAction[][] branch;

	InputTerminal[][] in;
	OutputTerminal[][] ackIn;

	OutputTerminal[][] intOut;
	InputTerminal[][] intAckOut;
		
	State[][] intDataState;
	State[][] intAckState;
	
	/** map from names to external terminals */
	Map<String, Terminal> nameToTerminalMap;
	
	// used by XML interface
	public DestinationHorn(String name, Parameters params) {
		super(name, null);
		
		int depthParam= hornAndFunnelDefaults.DEPTH_DEF;
		
		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				
				if (pName.equals(hornAndFunnelDefaults.DEPTH) 
						|| pName.equals(FleetGlobal.PARAMETER) ) {
					try {
						depthParam= Integer.parseInt(sValue);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ hornAndFunnelDefaults.DEPTH
								+ ": " + sValue
								+ ". Value is IGNORED, default value of "
								+ hornAndFunnelDefaults.DEPTH_DEF 
								+ " is used.");						
					}
				}
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");					
				}
				
			}
		} // if params != null		
		
		buildDestinationHorn(depthParam);
	}
	
	// used by XML interface
	public DestinationHorn(String name, int depth, Parameters params) {
		super(name, null);
		
		int depthParam= depth;;
		
		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				
				if (pName.equals(hornAndFunnelDefaults.DEPTH) ){
					// ignore depth
				}
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");					
				}
				
			}
		} // if params != null		
		
		buildDestinationHorn(depthParam);
	}	
	
	/**
	 * Make an instruction Horn. Has 2^depth outputs.
	 * @param name name
	 * @param d = depth the number of levels inside the horn
	 */
	public DestinationHorn(String name, int d) {
		super(name, null);
		
		buildDestinationHorn(d);

		
	} // DestinationHornConstructor
	
	
	protected void buildDestinationHorn(int d) {
		depth= d;
		numOutputs= 1 << depth;
		
		//prepare arrays for required components
		branch= new DestinationHornBranchAction[depth][];
		
		// inputs to branches
		in= new InputTerminal[depth][];
		ackIn= new OutputTerminal[depth][];
		
		// outputs from branches
		intOut= new OutputTerminal[depth][];
		intAckOut= new InputTerminal[depth][];	
		
		intDataState= new State[depth][];
		intAckState= new State[depth][];
		
		// mask used to steer data
		// set the bit that decides the direction to 1, the rest to 0
		// start by setting the high-order bit to 1
		int mask= (1 << (depth-1));
		// number of components per level
		int perLevel= 1;
		
		// create components for each level
		for (int level=0; level<depth; level++) {

			// prepare arrays for components in this level
			branch[level]= new DestinationHornBranchAction[perLevel];
			in[level] = new InputTerminal[perLevel];
			ackIn[level]= new OutputTerminal[perLevel];
			// outputs come in accordance with branch factor
			intOut[level]= new OutputTerminal[branchFactor*perLevel];
			intAckOut[level]= new InputTerminal[branchFactor*perLevel];
			// these states lead to the previous level
			intDataState[level]= new State[perLevel];
			intAckState[level]= new State[perLevel]; 
			
			
			
			// make components and connections for this level
			for (int i=0; i< perLevel; i++) {
				branch[level][i]= new DestinationHornBranchAction(
						FleetGlobal.DELIMITER+level+FleetGlobal.DELIMITER+i,
						mask);
				addMember(branch[level][i]);
				
				// make the input and ackIn
				in[level][i]= new InputTerminal(
						FleetGlobal.IN+FleetGlobal.DELIMITER
						+level+FleetGlobal.DELIMITER+i, this);
				ackIn[level][i]= new OutputTerminal(
						FleetGlobal.ACK_IN+FleetGlobal.DELIMITER
						+level+FleetGlobal.DELIMITER+i, this);
				// attach terminals to the input side of the action
				branch[level][i].attachIn(in[level][i], ackIn[level][i]);

				// now create terminals on the output side
				for (int j= 0; j<branchFactor; j++) {
					int index= i*branchFactor+j;
					intOut[level][index]= new OutputTerminal(
							FleetGlobal.OUT+FleetGlobal.DELIMITER
							+level+FleetGlobal.DELIMITER
							+i+FleetGlobal.DELIMITER+j, this);
					intAckOut[level][index]= new InputTerminal(
							FleetGlobal.ACK_OUT+FleetGlobal.DELIMITER
							+level+FleetGlobal.DELIMITER+i+FleetGlobal.DELIMITER+j,
							this);
					// atach the terminals
					branch[level][i].attachOut(intOut[level][index], 
							intAckOut[level][index], j);
				}
				
				// if this is not the first branch, establish the connection 
				if (level != 0) {
					// make the states
					intDataState[level][i]= new State(
							FleetGlobal.DATA_STATE+FleetGlobal.DELIMITER
							+level+FleetGlobal.DELIMITER+i, this);
					intAckState[level][i]= new State(
							FleetGlobal.ACK_STATE+FleetGlobal.DELIMITER
							+level+FleetGlobal.DELIMITER+i, this);
					intAckState[level][i].setInitiallyActive(null);
					
					// attach the terminals
					intDataState[level][i].attach(in[level][i]);
					// this one comes from the previous level
					intDataState[level][i].attach(intOut[level-1][i]);
					// an ack on this level
					intAckState[level][i].attach(ackIn[level][i]);
					// going a level back
					intAckState[level][i].attach(intAckOut[level-1][i]);
					
					// all of the internal ack states are active!
					// the state was created as an active one
					// intAckState[level][i].setActive(null);
					
				} // if level != 0 -> make internal states 
				  // and corresponding connections
				
			} // for i = 0..perLevel
			// next level branches according to the branchFactor
			perLevel= perLevel*branchFactor;
			// prepare the mask for the next level
			mask= mask >> 1;
		} // for level
		
		// set terminal names and establish name to terminal mapping
		// for all external terminals
		nameToTerminalMap= new HashMap<String, Terminal>();
		Terminal t;
		
		// get input - at ther root of the horn
		t= getIn();
		// set the name
		t.setAlias(FleetGlobal.IN);
		// establish the mapping
		nameToTerminalMap.put(FleetGlobal.IN, t);
		nameToTerminalMap.put(t.getBaseName(), t);
		
		// the same for the ackIn from root
		t= getAckIn();
		t.setAlias(FleetGlobal.ACK_IN);
		nameToTerminalMap.put(FleetGlobal.ACK_IN, t);
		nameToTerminalMap.put(t.getBaseName(), t);
		
		for (int i=0; i< numOutputs; i++) {
			// set the name and establish the mapping for the output side
			t= getOut(i);
			t.setAlias(FleetGlobal.OUT+i);
			nameToTerminalMap.put(FleetGlobal.OUT+i, t);
			nameToTerminalMap.put(t.getBaseName(), t);
			
			t= getAckOut(i);
			t.setAlias(FleetGlobal.ACK_OUT+i);
			nameToTerminalMap.put(FleetGlobal.ACK_OUT+i, t);
			nameToTerminalMap.put(t.getBaseName(), t);
			
		} // for
						
	}
	
	public InputTerminal getIn() {
		return in[0][0];
	}
	
	public OutputTerminal getAckIn() {
		return ackIn[0][0];
	}
	
	public OutputTerminal getOut(int n) {
		if ( 0 <= n && n < numOutputs)
			return intOut[depth-1][n];
		else
			throw new IndexOutOfBoundsException();
	}
	
	public InputTerminal getAckOut(int n) {
		if ( 0 <= n && n < numOutputs)
			return intAckOut[depth-1][n];
		else
			throw new IndexOutOfBoundsException();	
	}
	
	/**
	 * return a terminal from the external boundary
	 *  given the terminal or alias.
	 */
	public Terminal getTerminal(String name) {
		Object o= nameToTerminalMap.get(name);
		if (o instanceof Terminal)
			return (Terminal)o;
		else 
			return null;		
	} // getTerminal
	
} // DestinationHorn
